// win32ole.net.h
/*
 copyright(c) 2007 arton

 Usage of the works is permitted provided that this instrument is retained with the works, 
 so that any entity that uses the works is notified of this instrument. 

 DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY. 

 [2004, Fair License: rhid.com/fair] 
*/

#pragma once

using namespace System;
using namespace System::Runtime::InteropServices;
using namespace System::Collections::Generic;

namespace Win32olenet {

    public ref class Win32OleException
        : ApplicationException
    {
        HRESULT result;
    public:
        Win32OleException(String^ msg, Exception^ e)
            : ApplicationException(msg, e)
        {
        }
        Win32OleException(String^ msg)
            : ApplicationException(msg)
        {
        }
    };

    public ref class OleEnumerator 
        : System::Collections::IEnumerator, IDisposable
    {
        IEnumVARIANT* pEnumVariant;
        Object^ current;
    internal:
        OleEnumerator(IEnumVARIANT* p)
            : pEnumVariant(p)
        {
            current = nullptr;
        }
    public:
        virtual property Object^ Current 
        {
            Object^ get()
            {
                return current;
            }
        }
        virtual bool MoveNext();
        virtual void Reset();
        virtual ~OleEnumerator();
    };

    public ref class WIN32OLE : System::Collections::IEnumerable
    {
        LPDISPATCH pDisp;
        IDictionary<String^, DISPID>^ methods;
    public:
        WIN32OLE(String^ name)
        {
            pin_ptr<IDispatch> pvoid;
            CLSID clsid;
            LPOLESTR lpolestr = (LPOLESTR)Marshal::StringToHGlobalUni(name).ToPointer();
            HRESULT hr = CLSIDFromProgID(lpolestr, &clsid);
            Marshal::FreeHGlobal((IntPtr)lpolestr);
            if (hr == S_OK)
            {
                hr = CoCreateInstance(clsid, NULL,
                                    CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER,
                                    IID_IDispatch, (void**)&pvoid);

                if (hr != S_OK)
                {
                    throwException(hr);
                }
                else
                {
                    pDisp = (LPDISPATCH)pvoid;
                }
            }
            methods = gcnew Dictionary<String^, DISPID>();
        }

        Object^ Invoke(String^ name, array<Object^>^ args)
        {
            checkActivity();
            return invoke(name, args, 0, NULL, DISPATCH_METHOD | DISPATCH_PROPERTYGET);
        }

        Object^ InvokeWithNamedArguments(String^ name, array<Object^>^ args, Object^ namedArg)
        {
            checkActivity();
            if (System::Collections::Hashtable::typeid->IsInstanceOfType(namedArg) == false)
                throw gcnew Win32OleException("InvokeWithNamedArguments takes System::Collections::Hashtable but `" + namedArg->GetType()->ToString() + "' is given");

            System::Collections::Hashtable^ hash = (System::Collections::Hashtable^)namedArg;
            DISPID* names = NULL;
            array<Object^>^ realargs = args;
            if (hash->Count)
            {
                names = new DISPID[hash->Count + 1];
                *names = DISPID_UNKNOWN;
                LPOLESTR* polenames = new LPOLESTR[hash->Count + 1];
                *polenames = (LPOLESTR)Marshal::StringToHGlobalUni(name).ToPointer();
                realargs = gcnew array<Object^>(hash->Count + args->Length);
                System::Collections::IDictionaryEnumerator^ dic = hash->GetEnumerator();
                for (int i = 0; i < hash->Count; i++)
                {
                    if (!dic->MoveNext()) break;
                    *(names + 1 + i) = DISPID_UNKNOWN;
                    *(polenames + 1 + i) = (LPOLESTR)Marshal::StringToHGlobalUni(dic->Key->ToString()).ToPointer();
                    realargs[i] = dic->Value;
                }
                HRESULT hr = pDisp->GetIDsOfNames(IID_NULL, polenames, hash->Count + 1, LOCALE_SYSTEM_DEFAULT, names);
                for (int i = 0; i < hash->Count + 1; i++)
                    Marshal::FreeHGlobal((IntPtr)*(polenames + i));
                delete[] polenames;
                if (hr != S_OK)
                    throw gcnew Win32OleException("failed to get named argument info: `" + name + "'", createException(hr));
                for (int i = 0; i < args->Length; i++)
                {
                    realargs[hash->Count + i] = args[i];
                }
                if (!methods->ContainsKey(name))
                {
                    methods->Add(name, names[0]);
                }
            }
            try
            {
                Object^ o = invoke(name, realargs, hash->Count, &names[1], DISPATCH_METHOD | DISPATCH_PROPERTYGET);
                delete[] names;
                return o;
            }
            catch (Exception^ e)
            {
                delete[] names;
                throw e;
            }
        }

        Object^ SetProperty(String^ name, array<Object^>^ args)
        {
            checkActivity();
            DISPID dispid(DISPID_PROPERTYPUT);
            return invoke(name, args, 1, &dispid, DISPATCH_PROPERTYPUT);
        }

        virtual System::Collections::IEnumerator^ GetEnumerator()
        {
            checkActivity();
            unsigned int argErr(0);
            ::EXCEPINFO einfo;
            memset(&einfo, 0, sizeof(einfo));
            ::DISPPARAMS params = { NULL, NULL, 0, 0 };
            ::VARIANT result;
            VariantInit(&result);
            HRESULT hr = pDisp->Invoke(DISPID_NEWENUM, IID_NULL, LOCALE_SYSTEM_DEFAULT,
                DISPATCH_METHOD | DISPATCH_PROPERTYGET, &params, &result, &einfo, &argErr);
            if (hr == S_OK)
            {
                pin_ptr<IEnumVARIANT> pEnum = (IEnumVARIANT*)NULL;
                hr = result.punkVal->QueryInterface(IID_IEnumVARIANT, (void**)&pEnum);
                VariantClear(&result);
                if (hr == S_OK)
                    return gcnew OleEnumerator(pEnum);
            }
            throw gcnew Win32OleException("faield to get IEnum Interface", createException(hr));
        }

        int Release()
        {
            int n = 0;
            if (pDisp)
                n = pDisp->Release();
            pDisp = NULL;
            return n;
        }

        int ReferenceCount()
        {
            if (pDisp)
            {
                pDisp->AddRef();
                return pDisp->Release();
            }
            return 0;
        }

        virtual ~WIN32OLE() 
        {
            if (pDisp)
                pDisp->Release();
        }
    protected:
        WIN32OLE(IDispatch* p)
        {
            pDisp = p;
            methods = gcnew Dictionary<String^, DISPID>();
        }
        Object^ invoke(String^ name, array<Object^>^ args, int cNames, DISPID* pNames, WORD wFlag)
        {
            DISPID id;
            if (methods->ContainsKey(name))
            {
                id = methods->default[name];
            }
            else
            {
                LPOLESTR szName = (LPOLESTR)Marshal::StringToHGlobalUni(name).ToPointer();
                HRESULT hr = pDisp->GetIDsOfNames(IID_NULL, &szName, 1, LOCALE_SYSTEM_DEFAULT, &id);
                Marshal::FreeHGlobal((IntPtr)szName);
                if (hr != S_OK)
                {
                    throwException(hr);
                }
                methods->Add(name, id);
            }
            ::EXCEPINFO excepinfo;
            memset(&excepinfo, 0, sizeof(excepinfo));
            UINT argerr(0);
            ::DISPPARAMS params = {NULL, pNames, args->Length, cNames};
            if (args->Length > 0)
            {
                params.rgvarg = new ::VARIANT[args->Length];
                for (int i = 0; i < cNames; i++)
                {
                    setVariant(args[i], params.rgvarg + i);
                }
                for (int i = 0; i < args->Length - cNames; i++)
                {
                    setVariant(args[i + cNames], params.rgvarg + args->Length - i - 1);
                }
            }
            ::VARIANT result;
            VariantInit(&result);
            HRESULT hr = pDisp->Invoke(id, IID_NULL, LOCALE_SYSTEM_DEFAULT, wFlag,
                &params, &result, &excepinfo, &argerr);
            if (args->Length > 0)
            {
                for (int i = 0; i < args->Length; i++)
                {
                    VariantClear(params.rgvarg + i);
                }
                delete[] params.rgvarg;
            }
            if (hr != S_OK)
            {
                throwException(hr);
            }
            return getObject(result);
        }

        void checkActivity()
        {
            if (!pDisp)
            {
                throw gcnew Win32OleException("object was already disconnected");
            }
        }
    public:
        void throwException(HRESULT hr)
        {
            throw createException(hr);
        }
        static void setVariant(Object^ obj, VARIANT* pvar)
        {
            if (WIN32OLE::typeid->IsInstanceOfType(obj))
            {
                pvar->vt = VT_DISPATCH;
                pvar->pdispVal = ((WIN32OLE^)obj)->pDisp;
                pvar->pdispVal->AddRef();
            }
            else
            {
                Marshal::GetNativeVariantForObject(obj, (IntPtr)pvar);
            }
        }
        static Object^ getObject(VARIANT& v)
        {
            if ((V_VT(&v) & ~VT_BYREF) == VT_DISPATCH)
            {
                return gcnew WIN32OLE((V_ISBYREF(&v)) ? *V_DISPATCHREF(&v) : V_DISPATCH(&v));
            }
            Object^ o = Marshal::GetObjectForNativeVariant((IntPtr)&v);
            VariantClear(&v);
            return o;
        }
        Exception^ createException(HRESULT hr)
        {
            pin_ptr<IErrorInfo> peinfo;
            Exception^ excep;
            if (pDisp->QueryInterface(IID_IErrorInfo, (void**)&peinfo) == S_OK)
            {
                excep = Marshal::GetExceptionForHR((int)hr, (IntPtr)peinfo);
                peinfo->Release();
            }
            else
            {
                excep = Marshal::GetExceptionForHR((int)hr);
            }
            return excep;
        }
        
    };
}
